@@ -1,4 +1,4 @@
-/* tftp-hpa: $Id: tftpd.c,v 1.11 2001/04/23 22:50:52 hpa Exp $ */
+/* tftp-hpa: $Id: tftpd.c,v 1.17 2001/07/10 22:05:21 hpa Exp $ */
 
 /* $OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $	*/
 
@@ -41,7 +41,7 @@ static const char *copyright =
  All rights reserved.\n";
 /*static char sccsid[] = "from: @(#)tftpd.c	5.13 (Berkeley) 2/26/91";*/
 /*static char rcsid[] = "$OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $: tftpd.c,v 1.6 1997/02/16 23:49:21 deraadt Exp $";*/
-static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.11 2001/04/23 22:50:52 hpa Exp $";
+static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.17 2001/07/10 22:05:21 hpa Exp $";
 #endif /* not lint */
 
 /*
@@ -77,6 +77,7 @@ static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.11 2001/04/23 22:50:52 hpa
 #include "../config.h"
 #include "tftpsubs.h"
 #include "recvfrom.h"
+#include "remap.h"
 
 #ifdef HAVE_TCPWRAPPERS
 #include <tcpd.h>
@@ -126,7 +127,10 @@ char	**dirs;
 int	secure = 0;
 int	cancreate = 0;
 
+int verbosity = 0;
+
 struct formats;
+static struct rule *rewrite_rules = NULL;
 
 int tftp(struct tftphdr *, int);
 void nak(int);
@@ -150,12 +154,18 @@ struct options {
         { NULL,         NULL }
 };
 
+/* Simple handler for SIGHUP */
+static volatile sig_atomic_t caught_sighup = 0;
+static void handle_sighup(int sig)
+{
+  caught_sighup = 1;
+}
 
 
 static void
 usage(void)
 {
-	syslog(LOG_ERR, "Usage: %s [-c] [-u user] [-t timeout] [-r option...] [-s] [directory ...]",
+	syslog(LOG_ERR, "Usage: %s [-vc][-m mappings][-u user][-t timeout][-r option...] [-s] [directory ...]",
 	       __progname);
 	exit(1);
 }
@@ -180,7 +190,7 @@ main(int argc, char **argv)
 
 	openlog(__progname, LOG_PID | LOG_NDELAY, LOG_DAEMON);
 
-	while ((c = getopt(argc, argv, "csu:r:t:")) != -1)
+	while ((c = getopt(argc, argv, "csvu:r:t:m:")) != -1)
 		switch (c) {
 		case 'c':
 			cancreate = 1;
@@ -206,10 +216,30 @@ main(int argc, char **argv)
 			exit(1);
 		  }
 		  break;
-
+#ifdef WITH_REGEX
+		case 'm':
+		  {
+		    FILE *f;
+		    if ( rewrite_rules ) {
+		      syslog(LOG_ERR, "Multiple -m options");
+		      exit(1);
+		    }
+		    f = fopen(optarg, "rt");
+		    if ( !f ) {
+		      syslog(LOG_ERR, "Cannot open map file: %s: %m", optarg);
+		      exit(1);
+		    }
+		    rewrite_rules = parserulefile(f);
+		    fclose(f);
+		  }
+		  break;
+#endif
+		case 'v':
+		  verbosity++;
+		  break;
 		default:
-			usage();
-			break;
+		  usage();
+		  break;
 		}
 
 	for (; optind != argc; optind++) {
@@ -255,18 +285,29 @@ main(int argc, char **argv)
 	/* This means we don't want to wait() for children */
 	bsd_signal(SIGCHLD, SIG_IGN);
 
+	/* Take SIGHUP and use it to set a variable.  This
+	   is polled synchronously to make sure we don't
+	   lose packets as a result. */
+	bsd_signal(SIGHUP, handle_sighup);
+
 	do {
 	  fd_set readset;
 	  struct timeval tv_timeout;
-	  
+	  int rv;
 
 	  FD_ZERO(&readset);
 	  FD_SET(fd, &readset);
 	  tv_timeout.tv_sec = timeout;
 	  tv_timeout.tv_usec = 0;
 
-	  if ( select(fd+1, &readset, NULL, NULL, &tv_timeout) == 0 )
-	    exit(0);		/* Timeout, return to inetd */
+	  if ( caught_sighup )
+	    exit(0);		/* Return to inetd for respawn */
+
+	  rv = select(fd+1, &readset, NULL, NULL, &tv_timeout);
+	  if ( rv == -1 && errno == EINTR )
+	    continue;		/* Signal caught, reloop */
+	  if ( rv <= 0 )
+	    exit(0);		/* Timeout or error, return to inetd */
 
 	  fromlen = sizeof (from);
 	  n = myrecvfrom(fd, buf, sizeof (buf), 0,
@@ -300,9 +341,8 @@ main(int argc, char **argv)
 #endif
 
 	/*
-	 * Now that we have read the message out of the UDP
-	 * socket, we fork and go back to listening to the
-	 * socket.
+	 * Now that we have read the request packet from the UDP
+	 * socket, we fork and go back to listening to the socket.
 	 */
 	  pid = fork();
 	  if (pid < 0) {
@@ -369,20 +409,22 @@ main(int argc, char **argv)
 	exit(0);
 }
 
+char   *rewrite_access(char *, int);
 int	validate_access(char *, int, struct formats *);
 void	sendfile(struct formats *, struct tftphdr *, int);
 void	recvfile(struct formats *, struct tftphdr *, int);
 
 struct formats {
 	char	*f_mode;
+  	char	*(*f_rewrite)(char *, int);
 	int	(*f_validate)(char *, int, struct formats *);
 	void	(*f_send)(struct formats *, struct tftphdr *, int);
 	void	(*f_recv)(struct formats *, struct tftphdr *, int);
 	int	f_convert;
 } formats[] = {
-	{ "netascii",	validate_access,	sendfile,	recvfile, 1 },
-	{ "octet",	validate_access,	sendfile,	recvfile, 0 },
-	{ NULL, NULL, NULL, NULL, 0 }
+	{ "netascii",   rewrite_access, validate_access, sendfile, recvfile, 1 },
+	{ "octet",	rewrite_access, validate_access, sendfile, recvfile, 0 },
+	{ NULL, NULL, NULL, NULL, NULL, 0 }
 };
 
 /*
@@ -394,6 +436,7 @@ tftp(struct tftphdr *tp, int size)
 	char *cp;
 	int argn, ecode;
 	struct formats *pf = NULL;
+	char *origfilename;
 	char *filename, *mode = NULL;
 
         char *val = NULL, *opt = NULL;
@@ -401,7 +444,7 @@ tftp(struct tftphdr *tp, int size)
 
         ((struct tftphdr *)ackbuf)->th_opcode = ntohs(OACK);
 
-	filename = cp = tp->th_stuff;
+	origfilename = cp = tp->th_stuff;
 	argn = 0;
 
 	while ( cp < buf + size && *cp ) {
@@ -428,6 +471,20 @@ tftp(struct tftphdr *tp, int size)
 		       nak(EBADOP);
 		       exit(0);
 		  }
+		  if ( !(filename = (*pf->f_rewrite)(origfilename, tp->th_opcode)) ) {
+		       nak(EACCES); /* File denied by mapping rule */
+		       exit(0);
+		  }
+		  if ( verbosity >= 1 ) {
+		    if ( filename == origfilename || !strcmp(filename, origfilename) )
+		      syslog(LOG_NOTICE, "%s from %s filename %s\n",
+			     tp->th_opcode == WRQ ? "WRQ" : "RRQ",
+			     inet_ntoa(from.sin_addr), filename);
+		    else
+		      syslog(LOG_NOTICE, "%s from %s filename %s remapped to %s\n",
+			     tp->th_opcode == WRQ ? "WRQ" : "RRQ",
+			     inet_ntoa(from.sin_addr), origfilename, filename);
+		  }		   
 		  ecode = (*pf->f_validate)(filename, tp->th_opcode, pf);
 		  if (ecode) {
 		       nak(ecode);
@@ -596,9 +653,22 @@ do_opt(char *opt, char *val, char **ap)
      return;
 }
 
+/*
+ * Modify the filename, if applicable.  If it returns NULL, deny the access.
+ */
+char *
+rewrite_access(char *filename, int mode)
+{
+#ifdef WITH_REGEX
+  if ( rewrite_rules ) {
+    char *newname = rewrite_string(filename, rewrite_rules, mode != RRQ);
+    filename = newname;
+  }
+#endif
+  return filename;
+}
 
 FILE *file;
-
 /*
  * Validate file access.  Since we
  * have no uid or gid, for now require
@@ -687,6 +757,7 @@ validate_access(char *filename, int mode, struct formats *pf)
 int	timeout;
 sigjmp_buf	timeoutbuf;
 
+/* Handle timeout signal */
 void
 timer(int sig)
 {
@@ -795,6 +866,7 @@ abort:
 	(void) fclose(file);
 }
 
+/* Bail out signal handler */
 void
 justquit(int sig)
 {
@@ -928,6 +1000,12 @@ nak(int error)
 	length = strlen(pe->e_msg);
 	tp->th_msg[length] = '\0';
 	length += 5;
+
+	if ( verbosity >= 2 ) {
+	  syslog(LOG_INFO, "sending NAK (%d, %s) to %s",
+		 error, tp->th_msg, inet_ntoa(from.sin_addr));
+	}
+
 	if (send(peer, buf, length, 0) != length)
 		syslog(LOG_ERR, "nak: %m");
 }
